/**
 * Copyright Marcus Britanicus <marcusbritanicus@gmail.com>
 *
 * This file is a part of LibDesQ (https://gitlab.com/DesQ/libdesq)
 * This library contains various  Core and Gui classes which are used
 * through out the DesQ Project.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * at your option, any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 **/

#include <unistd.h>
#include <QStorageInfo>

#include "Xdg.hpp"
#include "Utils.hpp"

#include "XdgDesktopImpl.hpp"
#include "XdgMimeAppsImpl.hpp"

/**
 * DesQ::XDG Helper Functions
 *  These helper functions will return XDG standard locations
 */
QString DesQ::XDG::homeDir() {
    QString home( qgetenv( "HOME" ) );

    if ( home.count() and DesQ::Utils::exists( home ) ) {
        return home + (home.endsWith( "/" ) ? "" : "/");
    }

    return QDir::homePath() + "/";
}


QString DesQ::XDG::xdgCacheHome() {
    QString dir( qgetenv( "XDG_CACHE_HOME" ) );

    if ( dir.count() and DesQ::Utils::exists( dir ) ) {
        return dir + (dir.endsWith( "/" ) ? "" : "/");
    }

    return QDir::home().filePath( ".cache/" );
}


QString DesQ::XDG::xdgConfigHome() {
    QString dir( qgetenv( "XDG_CONFIG_HOME" ) );

    if ( dir.count() and DesQ::Utils::exists( dir ) ) {
        return dir + (dir.endsWith( "/" ) ? "" : "/");
    }

    return QDir::home().filePath( ".config/" );
}


QString DesQ::XDG::xdgDataHome() {
    QString dir( qgetenv( "XDG_DATA_HOME" ) );

    if ( dir.count() and DesQ::Utils::exists( dir ) ) {
        return dir + (dir.endsWith( "/" ) ? "" : "/");
    }

    return QDir::home().filePath( ".local/share/" );
}


QString DesQ::XDG::xdgStateHome() {
    QString dir( qgetenv( "XDG_STATE_HOME" ) );

    if ( dir.count() and DesQ::Utils::exists( dir ) ) {
        return dir + (dir.endsWith( "/" ) ? "" : "/");
    }

    return QDir::home().filePath( ".local/state/" );
}


QStringList DesQ::XDG::xdgConfigDirs() {
    QString dirs = qgetenv( "XDG_CONFIG_DIRS" );

    if ( dirs.count() ) {
        return dirs.split( ":", Qt::SkipEmptyParts );
    }

    return QStringList( { "/etc/xdg/" } );
}


QStringList DesQ::XDG::xdgDataDirs() {
    QString dirs = qgetenv( "XDG_DATA_DIRS" );

    if ( dirs.count() ) {
        return dirs.split( ":", Qt::SkipEmptyParts );
    }

    return QStringList( { "/usr/local/share/", "/usr/share/" } );
}


QString DesQ::XDG::xdgDesktopDir() {
    QString dir( qgetenv( "XDG_DESKTOP_DIR" ) );

    if ( dir.count() and DesQ::Utils::exists( dir ) ) {
        return dir + (dir.endsWith( "/" ) ? "" : "/");
    }

    return QDir::homePath();
}


QString DesQ::XDG::xdgDocumentsDir() {
    QString dir( qgetenv( "XDG_DOCUMENTS_DIR" ) );

    if ( dir.count() and DesQ::Utils::exists( dir ) ) {
        return dir + (dir.endsWith( "/" ) ? "" : "/");
    }

    return QDir::homePath();
}


QString DesQ::XDG::xdgDownloadDir() {
    QString dir( qgetenv( "XDG_DOWNLOAD_DIR" ) );

    if ( dir.count() and DesQ::Utils::exists( dir ) ) {
        return dir + (dir.endsWith( "/" ) ? "" : "/");
    }

    return QDir::homePath();
}


QString DesQ::XDG::xdgMusicDir() {
    QString dir( qgetenv( "XDG_MUSIC_DIR" ) );

    if ( dir.count() and DesQ::Utils::exists( dir ) ) {
        return dir + (dir.endsWith( "/" ) ? "" : "/");
    }

    return QDir::homePath();
}


QString DesQ::XDG::xdgPicturesDir() {
    QString dir( qgetenv( "XDG_PICTURES_DIR" ) );

    if ( dir.count() and DesQ::Utils::exists( dir ) ) {
        return dir + (dir.endsWith( "/" ) ? "" : "/");
    }

    return QDir::homePath();
}


QString DesQ::XDG::xdgPublicshareDir() {
    QString dir( qgetenv( "XDG_PUBLICSHARE_DIR" ) );

    if ( dir.count() and DesQ::Utils::exists( dir ) ) {
        return dir + (dir.endsWith( "/" ) ? "" : "/");
    }

    return QDir::homePath();
}


QString DesQ::XDG::xdgTemplatesDir() {
    QString dir( qgetenv( "XDG_TEMPLATES_DIR" ) );

    if ( dir.count() and DesQ::Utils::exists( dir ) ) {
        return dir + (dir.endsWith( "/" ) ? "" : "/");
    }

    return QDir::homePath();
}


QString DesQ::XDG::xdgVideosDir() {
    QString dir( qgetenv( "XDG_VIDEOS_DIR" ) );

    if ( dir.count() and DesQ::Utils::exists( dir ) ) {
        return dir + (dir.endsWith( "/" ) ? "" : "/");
    }

    return QDir::homePath();
}


QString DesQ::XDG::xdgRuntimeDir() {
    /* Get it from the env variable */
    QString dir( qgetenv( "XDG_RUNTIME_DIR" ) );

    /* Variable not set */
    if ( not dir.count() ) {
        dir = QString( "/run/user/%1" ).arg( getuid() );

        /* We will try to create /run/$UID */
        if ( DesQ::Utils::mkpath( dir, 0700 ) ) {
            return dir + (dir.endsWith( "/" ) ? "" : "/");
        }
    }

    /* So the variable is set, but does not exist */
    else if ( not DesQ::Utils::exists( dir ) ) {
        /* Created successfully */
        if ( DesQ::Utils::mkpath( dir, 0700 ) ) {
            return dir + (dir.endsWith( "/" ) ? "" : "/");
        }
    }

    /** We failed with /run/user/$USER. Time for /tmp/$USER */
    dir = QString( "/tmp/%1/" ).arg( getuid() );

    if ( DesQ::Utils::mkpath( dir, 0700 ) ) {
        return dir;
    }

    return QString( "/tmp/" );
}


/**
 * DesQ::DesktopFile
 *  This class will parse, and optionally run, an app given its desktop file.
 *  Add/Remove the desktop urls from memory when an app is installed/uninstalled.
 *  It will also list the URLs for a given mimetype
 *  Provide the default app desktop url for a given mimetype.
 */

DesQ::DesktopFile::DesktopFile( QString filename ) {
    if ( filename.isEmpty() ) {
        return;
    }

    mFileUrl     = desktopPathForName( filename );
    mDesktopName = DesQ::Utils::baseName( mFileUrl );

    if ( mFileUrl.startsWith( "/usr/share/applications" ) ) {
        mRank = 1;
    }

    else if ( mFileUrl.startsWith( "/usr/share/applications/kde4" ) ) {
        mRank = 2;
    }

    else if ( mFileUrl.startsWith( "/usr/share/gnome/applications" ) ) {
        mRank = 3;
    }

    else if ( mFileUrl.startsWith( "/usr/local/share/applications" ) ) {
        mRank = 4;
    }

    else if ( mFileUrl.startsWith( QDir::home().filePath( ".local/share/applications" ) ) ) {
        mRank = 5;
    }

    QMap<QString, QVariant> desktopMap = parseDesktop( mFileUrl );

    if ( desktopMap.count() == 0 ) {
        return;
    }

    mName        = desktopMap.value( "Name" ).toString();
    mGenericName = desktopMap.value( "GenericName", mName ).toString();
    mDescription = desktopMap.value( "Comment", "No description available" ).toString();
    mCommand     = desktopMap.value( "Exec" ).toString();
    mExec        = desktopMap.value( "TryExec", mCommand.split( " ", Qt::SkipEmptyParts ).value( 0 ) ).toString();

    if ( not DesQ::Utils::exists( mExec ) ) {
        for ( QString path: QString::fromLocal8Bit( qgetenv( "PATH" ) ).split( ":", Qt::SkipEmptyParts ) ) {
            if ( DesQ::Utils::exists( path + "/" + mExec ) ) {
                mExec = path + "/" + mExec;
                break;
            }
        }
    }

    mIcon = desktopMap.value( "Icon" ).toString();

    QStringList args = mCommand.split( " " );

    for ( QString arg: args ) {
        if ( arg == "%f" or arg == "%u" ) {
            mMultiArgs = false;
            mTakesArgs = true;
        }

        else if ( arg == "%F" or arg == "%U" ) {
            mMultiArgs = true;
            mTakesArgs = true;
        }
    }

    mMimeTypes  = desktopMap.value( "MimeType" ).toStringList();
    mCategories = desktopMap.value( "Categories" ).toStringList();

    getCategory();

    mVisible       = not desktopMap.value( "NoDisplay", false ).toBool();
    mRunInTerminal = desktopMap.value( "Terminal", false ).toBool();

    if ( desktopMap.value( "Type" ).toString() == "Application" ) {
        mType = Application;
    }

    else if ( desktopMap.value( "Type" ).toString() == "Link" ) {
        mType = Link;
    }

    else if ( desktopMap.value( "Type" ).toString() == "Directory" ) {
        mType = Directory;
    }

    /** Update actions */
    mActionNames = desktopMap.value( "Actions" ).toStringList();
    for ( QString act: mActionNames ) {
        mActions[ act ] = desktopMap[ act ].toStringList();
    }

    if ( mName.count() and mCommand.count() ) {
        if ( not DesQ::Utils::exists( mExec ) ) {
            mValid = false;
        }

        else {
            mValid = true;
        }
    }
}


bool DesQ::DesktopFile::startApplication( QStringList args, QString action, QMap<QString, QString> environ ) {
    if ( not mValid ) {
        return false;
    }

    QProcess proc;

    QStringList execList;

    /** Parse the exec */
    QStringList execArgs;

    if ( action.count() and mActionNames.contains( action ) ) {
        execArgs = mActions[ action ][ 2 ].split( " " );
    }

    else {
        execArgs = mCommand.split( " " );
    }

    QStringList parsedArgs;

    for ( QString arg: execArgs ) {
        if ( arg == "%f" or arg == "%u" ) {
            parsedArgs << "<#DESQARG-FILE#>";
        }

        else if ( arg == "%F" or arg == "%U" ) {
            parsedArgs << "<#DESQARG-FILES#>";
        }

        else if ( arg == "%i" ) {
            if ( !mIcon.isEmpty() ) {
                parsedArgs << "--icon" << mIcon;
            }
        }

        else if ( arg == "%c" ) {
            parsedArgs << mName;
        }

        else if ( arg == "%k" ) {
            parsedArgs << QUrl( mFileUrl ).toLocalFile();
        }

        else {
            parsedArgs << arg;
        }
    }

    /* If this is a commandline app */
    if ( mRunInTerminal ) {
        execList << QFileInfo( "/etc/alternatives/x-terminal-emulator" ).symLinkTarget() << "-e";
    }

    execList << parsedArgs;
    QString exec = execList.takeFirst();

    QProcess appProc;
    appProc.setProgram( exec );
    appProc.setWorkingDirectory( XDG::homeDir() );

    QProcessEnvironment pe( QProcessEnvironment::systemEnvironment() );
    for ( QString key: environ.keys() ) {
        pe.insert( key, environ.value( key ) );
    }

    appProc.setProcessEnvironment( pe );

    if ( not args.count() ) {
        execList.removeAll( "<#DESQARG-FILES#>" );
        execList.removeAll( "<#DESQARG-FILE#>" );

        appProc.setArguments( execList );
        return appProc.startDetached();
    }

    QStringList argList;

    if ( mTakesArgs ) {
        if ( mMultiArgs ) {
            for ( QString exeArg: execList ) {
                if ( exeArg == "<#DESQARG-FILES#>" ) {
                    if ( args.count() ) {
                        argList << args;
                    }
                }

                else {
                    argList << exeArg;
                }
            }
        }

        else {
            int idx = exec.indexOf( "<#DESQARG-FILE#>" );
            argList << execList;
            argList.removeAt( idx );

            if ( args.count() ) {
                argList.insert( idx, args.takeAt( 0 ) );
                argList << args;
            }
        }
    }

    else {
        argList << execList;

        if ( args.count() ) {
            argList << args;
        }
    }

    appProc.setArguments( argList );

    return appProc.startDetached();
}


QString DesQ::DesktopFile::desktopName() const {
    return mDesktopName;
}


QString DesQ::DesktopFile::name() const {
    return mName;
}


QString DesQ::DesktopFile::genericName() const {
    return mGenericName;
}


QString DesQ::DesktopFile::description() const {
    return mDescription;
}


QString DesQ::DesktopFile::exec() const {
    return mExec;
}


QString DesQ::DesktopFile::command() const {
    return mCommand;
}


QString DesQ::DesktopFile::icon() const {
    return mIcon;
}


QString DesQ::DesktopFile::category() const {
    return mCategory;
}


QStringList DesQ::DesktopFile::mimeTypes() const {
    return mMimeTypes;
}


QStringList DesQ::DesktopFile::categories() const {
    return mCategories;
}


QStringList DesQ::DesktopFile::actions() const {
    return mActionNames;
}


QStringList DesQ::DesktopFile::action( QString id ) const {
    return mActions[ id ];
}


int DesQ::DesktopFile::type() const {
    return mType;
}


int DesQ::DesktopFile::rank() const {
    return mRank;
}


bool DesQ::DesktopFile::visible() const {
    return mVisible;
}


bool DesQ::DesktopFile::runInTerminal() const {
    return mRunInTerminal;
}


bool DesQ::DesktopFile::multipleArgs() const {
    return mMultiArgs;
}


bool DesQ::DesktopFile::isValid() const {
    return mValid;
}


QString DesQ::DesktopFile::desktopFileUrl() const {
    return mFileUrl;
}


bool DesQ::DesktopFile::operator==( const DesQ::DesktopFile& other ) const {
    bool truth = true;

    truth &= (mCommand == other.command() );
    truth &= (mRank == other.rank() );

    return truth;
}


QString DesQ::DesktopFile::desktopPathForName( QString desktopName ) {
    if ( not desktopName.endsWith( ".desktop" ) ) {
        desktopName += ".desktop";
    }

    if ( DesQ::Utils::exists( desktopName ) ) {
        return desktopName;
    }

    /** Assuming desktopName is the name of the desktop */
    for ( QString appDirStr: DesQ::ApplicationManager::applicationPaths() ) {
        if ( DesQ::Utils::exists( appDirStr + "/" + desktopName ) ) {
            return appDirStr + "/" + desktopName;
        }
    }

    return QString();
}


void DesQ::DesktopFile::getCategory() {
    QStringList Accessories = QStringList() << "Utility" << "Utilities" << "Accessory" << "Accessories";
    QStringList Development = QStringList() << "Development";
    QStringList Education   = QStringList() << "Education";
    QStringList Games       = QStringList() << "Games" << "Game" << "ArcadeGame" << "StrategyGame" << "LogicGame";
    QStringList Graphics    = QStringList() << "Graphics";
    QStringList Internet    = QStringList() << "Network" << "Internet";
    QStringList Multimedia  = QStringList() << "Audio" << "Video" << "AudioVideo" << "Multimedia";
    QStringList Office      = QStringList() << "Office";
    QStringList ScienceMath = QStringList() << "Science" << "Math";
    QStringList Settings    = QStringList() << "Settings";
    QStringList System      = QStringList() << "System";
    QStringList Wine        = QStringList() << "Wine";

    /* If the file name and the command contains 'wine', then it's a wine application */
    if ( mFileUrl.contains( "wine", Qt::CaseInsensitive ) and mCommand.contains( "wine", Qt::CaseInsensitive ) ) {
        mCategories = QStringList( { "Wine" } );
        mCategory   = "Wine";
        return;
    }

    else {
        mCategory = "Uncategorized";
        foreach( QString cate, mCategories ) {
            if ( Accessories.contains( cate ) ) {
                mCategory = "Accessories";
            }

            else if ( Development.contains( cate ) ) {
                mCategory = "Development";
            }

            else if ( ScienceMath.contains( cate ) ) {
                mCategory = "Science & Math";
            }

            else if ( Education.contains( cate ) ) {
                mCategory = "Education";
            }

            else if ( Games.contains( cate ) ) {
                mCategory = "Games";
            }

            else if ( Office.contains( cate ) ) {
                mCategory = "Office";
            }

            else if ( Graphics.contains( cate ) ) {
                mCategory = "Graphics";
            }

            else if ( Internet.contains( cate ) ) {
                mCategory = "Internet";
            }

            else if ( Multimedia.contains( cate ) ) {
                mCategory = "Multimedia";
            }

            else if ( Settings.contains( cate ) ) {
                mCategory = "Settings";
            }

            else if ( System.contains( cate ) ) {
                mCategory = "System";
            }

            else if ( Wine.contains( cate ) ) {
                mCategory = "Wine";
            }

            else {
                continue;
            }
        }
    }
}


/**
 * DesQ::ApplicationManager
 *  This class will manage the desktops.
 *  Add/Remove the desktop urls from memory when an app is installed/uninstalled.
 *  It will also list the URLs for a given mimetype
 *  Provide the default app desktop url for a given mimetype.
 */

DesQ::ApplicationManager::ApplicationManager() : QObject() {
    parseDesktops();

    defaultWriter = new MimeAppsListParser( QDir::home().filePath( ".config/mimeapps.list" ) );
}


DesQ::AppsList DesQ::ApplicationManager::allApplications() {
    return mAllDesktops.values();
}


QStringList DesQ::ApplicationManager::appsForMimeType( QString mimeType ) {
    return mimeAppsHash.value( mimeType );
}


QString DesQ::ApplicationManager::defaultAppForMimeType( QString mimeType ) {
    QString defApp( mimeDefaultAppHash.value( mimeType ) );

    if ( defApp.isEmpty() ) {
        QStringList others( mimeAppsHash.value( mimeType ) );

        if ( others.count() == 1 ) {
            defApp = others[ 0 ];
        }
    }

    return defApp;
}


void DesQ::ApplicationManager::setDefaultAppForMimeType( QString mimeType, QString desktop ) {
    defaultWriter->setValue( "Default Applications", mimeType, desktop );

    mimeDefaultAppHash[ mimeType ] = desktop;
}


QStringList DesQ::ApplicationManager::applicationPaths() {
    return {
        QDir::home().filePath( ".local/share/applications/" ),
        "/usr/local/share/applications/",
        "/usr/share/applications/kde4/",
        "/usr/share/gnome/applications/",
        "/usr/share/applications/"
    };
}


void DesQ::ApplicationManager::parseDesktops() {
    QStringList paths( applicationPaths() );

    std::reverse( paths.begin(), paths.end() );

    for ( QString appDir: paths ) {
        QDirIterator it( appDir, { "*.desktop" }, QDir::Files );

        while ( it.hasNext() ) {
            DesQ::DesktopFile desktop( it.next() );

            if ( not desktop.isValid() ) {
                continue;
            }

            /** If it's not an application type desktop, ignore it */
            if ( desktop.type() != DesQ::DesktopFile::Type::Application ) {
                continue;
            }

            /**
             * If we have a higher range coming later on,
             * the current value will be replaced by the newer one.
             */
            mAllDesktops[ desktop.desktopName() ] = desktop;

            QStringList mimeTypes( desktop.mimeTypes() );
            QString     desktopName( desktop.desktopName() );

            /**
             * First, if any mimetype in mimeAppsHash lists this desktop,
             * but, the current desktop (higher rank) does not contain that
             * mimetype, then remove the association.
             */
            for ( QString mimeType: mimeAppsHash.keys() ) {
                QStringList desktops( mimeAppsHash[ mimeType ] );

                if ( desktops.contains( desktopName ) and not mimeType.contains( mimeType ) ) {
                    desktops.removeAll( desktopName );
                    mimeAppsHash[ mimeType ] = desktops;
                }
            }

            /** Update the has with mimetypes of current desktop */
            for ( QString mimeType: desktop.mimeTypes() ) {
                if ( not mimeAppsHash.contains( mimeType ) ) {
                    mimeAppsHash[ mimeType ] = QStringList();
                }

                /** Add to the list only if it does not exist */
                if ( not mimeAppsHash.value( mimeType ).contains( desktopName ) ) {
                    mimeAppsHash[ mimeType ] << desktopName;
                }
            }

            /** Handle Added/Removed Associations of this path's mimeapps.list */
            MimeAppsListParser parser( appDir + "/mimeapps.list" );

            /** Added Associations */
            for ( QString key: parser.keys( "Added Associations" ) ) {
                QVariant value = parser.value( "Added Associations", key );

                if ( not mimeAppsHash.contains( key ) ) {
                    mimeAppsHash[ key ] = QStringList();
                }

                QStringList desktops( mimeAppsHash[ key ] );
                for ( QString desktopName: value.toStringList() ) {
                    if ( DesQ::DesktopFile( desktopName ).isValid() and not desktops.contains( desktopName ) ) {
                        mimeAppsHash[ key ] << desktopName;
                    }
                }
            }

            /** Removed Associations */
            for ( QString key: parser.keys( "Removed Associations" ) ) {
                QVariant value = parser.value( "Removed Associations", key );

                QStringList desktops( mimeAppsHash.value( key ) );
                for ( QString desktopName: value.toStringList() ) {
                    desktops.removeAll( desktopName );
                    mimeAppsHash[ key ] = desktops;
                }
            }
        }
    }

    /** Parsing mimeapps.list
     *  1. Parse /usr/share/mimeapps.list
     *  2. Parse /usr/local/share/mimeapps.list (Will override /usr/share/)
     *  3. Parse ~/.local/share/mimeapps.list (Will override /usr/share/ and /usr/local/share)
     *
     *  First valid desktop will be the default
     */

    QStringList appPaths = {
        "/usr/share/applications/",
        "/usr/share/applications/kde4/",
        "/usr/share/gnome/applications/",
        "/usr/local/share/applications/",
        QDir::home().filePath( ".config/" ),
    };

    for ( QString path: appPaths ) {
        MimeAppsListParser parser( path + "/mimeapps.list" );
        for ( QString key: parser.keys( "Default Applications" ) ) {
            QVariant value = parser.value( "Default Applications", key );

            for ( QString desktopName: value.toStringList() ) {
                if ( DesQ::DesktopFile( desktopName ).isValid() ) {
                    mimeDefaultAppHash[ key ] = desktopName;
                    break;
                }
            }
        }
    }
}


uint qHash( const DesQ::DesktopFile& app ) {
    QString hashString;

    hashString += app.name();
    hashString += app.genericName();
    hashString += app.description();
    hashString += app.command();
    hashString += app.icon();
    hashString += app.mimeTypes().join( " " );
    hashString += app.categories().join( " " );

    return qChecksum( hashString.toLocal8Bit().data(), hashString.count() );
}
